//
//  peerguardnf (PeerGuard for Linux, BSD, and MacOS X systems)
//  Copyright (C) 2005 bitrot
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Id: Blocker_BSD.cpp,v 1.2 2005/03/27 07:38:10 nullcontext Exp $

#if defined(__APPLE__) || defined(__BSD__)

#include "PeerGuard.h"

#include <netinet/in_systm.h>
#include <sys/socket.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>

using namespace std;

namespace PG {

void* BlockerThread(void* block_list_)
{
   unsigned char buf[0xFFFF];
   BlockList* block_list = (BlockList*)block_list_;

   int sock = socket(PF_INET, SOCK_RAW, IPPROTO_DIVERT);

   struct timeval timeout = { 1, 0 };
   setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, &timeout, sizeof(timeout));
   setsockopt(sock, SOL_SOCKET, SO_SNDTIMEO, &timeout, sizeof(timeout));

   struct sockaddr_in server;

   server.sin_family = PF_INET;
   server.sin_port = htons(6137);
   server.sin_addr.s_addr = INADDR_ANY;

   if (bind(sock, (struct sockaddr*)&server, sizeof(server)) < 0)
   {
      *gLogStream << "bind error" << endl;
      exit(1);
   }

   for (int i = 0; i < 0xFFFF; i++)
   {
      if (gPrefs.allowedPorts[i])
         *gLogStream << "Allowing all traffic on port " << i << endl;
   }

   *gLogStream << "Blocking " << block_list->RangeCount() 
               << " ranges (" << block_list->IpCount()
               << " IP addresses)" << endl;

   // Main loop, read and block packets
   while (!gStopBlockerThread)
   {
      struct sockaddr_in client;
      int size, clisize = sizeof(client);
      size = recvfrom(sock, buf, sizeof(buf), 0, (struct sockaddr *)&client, &clisize);
      if (size < 0)
      {
         if (errno == EAGAIN)
            continue;

         *gLogStream << "recvfrom failed" << endl;
         exit(1);
      }

      bool verdict = true;

      if (size >= (int)sizeof(struct ip))
      {
         struct ip* iphdr = (struct ip*)buf;
         Range find;

         // Check allowed ports
         if (iphdr->ip_p == IPPROTO_TCP)
         {
            struct tcphdr* tcphdr = (struct tcphdr*)&buf[sizeof(struct ip)];
            if (gPrefs.allowedPorts[ntohs(tcphdr->th_sport)] || gPrefs.allowedPorts[ntohs(tcphdr->th_dport)])
               goto ProcessPacket;
         }
         else if (iphdr->ip_p == IPPROTO_UDP)
         {
            struct udphdr* udphdr = (struct udphdr*)&buf[sizeof(struct ip)];
            if (gPrefs.allowedPorts[ntohs(udphdr->uh_sport)] || gPrefs.allowedPorts[ntohs(udphdr->uh_dport)])
               goto ProcessPacket;
         }

         // Check block list
         if (block_list->Find(&find, iphdr->ip_src.s_addr) || block_list->Find(&find, iphdr->ip_dst.s_addr))
         {
            verdict = false;
            unsigned char* s_ip = (unsigned char*)&iphdr->ip_src.s_addr;
            unsigned char* d_ip = (unsigned char*)&iphdr->ip_dst.s_addr;

            *gLogStream << "dropped src: "
               << (int)s_ip[0] << "." << (int)s_ip[1] << "." << (int)s_ip[2] << "." << (int)s_ip[3]
               << " dst: "
               << (int)d_ip[0] << "." << (int)d_ip[1] << "." << (int)d_ip[2] << "." << (int)d_ip[3]
               << " (" << find.rule << ")" << endl;

            (*block_list)++;
         }
      }

ProcessPacket:
      if (verdict && sendto(sock, buf, size, 0, (struct sockaddr *)&client, clisize) != size)
      {
         *gLogStream << "Failed to write packet" << endl;
         exit(1);
      }
   }

   return 0;
}

} // namespace PG

#endif // defined(__APPLE__) || defined(__BSD__)

