//
//  peerguardnf (PeerGuard for Linux, BSD, and MacOS X systems)
//  Copyright (C) 2005 bitrot
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Id: Blocker_Linux.cpp,v 1.6 2005/03/27 07:38:10 nullcontext Exp $

#if !defined(__APPLE__) && !defined(__BSD__)

#include "PeerGuard.h"

#include <linux/netfilter.h>
extern "C" {
#include <libipq.h>
}

#include <sys/socket.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>

#define FIFO_NAME "/tmp/PG_fifo"
#define BUFFER_SIZE PIPE_BUF
using namespace std;

namespace PG {
// Globals
int num_checked=0, num_blocked=0;

void *talktoUI(void* arg)
{
	int PG_pipe=0, my_response=0,num_sent=0;
	char buffer[BUFFER_SIZE + 1];
	my_response = mkfifo(FIFO_NAME,0777);
   	chmod(FIFO_NAME, 0777);
        
	while(!gStopBlockerThread){
		PG_pipe = open(FIFO_NAME, O_WRONLY);	
		
		if(PG_pipe != -1){
			num_sent++;
			sprintf(buffer, "PB%d", num_blocked);
			my_response = write(PG_pipe, buffer, BUFFER_SIZE);
			//if(my_response == -1){
			//	fprintf(stderr,"Write error on pipe: %s %d\n", buffer, num_sent);
			//}
			sprintf(buffer, "PC%d", num_checked);
			my_response = write(PG_pipe, buffer, BUFFER_SIZE);
			//if(my_response == -1){
			//	fprintf(stderr,"Write error on pipe: %s %d\n", buffer, num_sent);
			//}
		}
		else{
			fprintf(stderr, "ERROR OPENING PIPE %d", num_sent);
		}
		(void)close(PG_pipe); 
	}
	pthread_exit(NULL);
}

void* BlockerThread(void* block_list_)
{
   unsigned char buf[0xFFFF];
   BlockList* block_list = (BlockList*)block_list_;
   int thread_response=0;
 
   pthread_t a_thread;
   
   for (int i = 0; i < 0xFFFF; i++)
   {
      if (gPrefs.allowedPorts[i])
         *gLogStream << "Allowing all traffic on port " << i << endl;
   }

   *gLogStream << "Blocking " << block_list->RangeCount() 
               << " ranges (" << block_list->IpCount()
               << " IP addresses)" << endl;

   struct ipq_handle* hIpq = ipq_create_handle(0, PF_INET);
   if (!hIpq)
   {
      *gLogStream << "ipq_create_handle failed; is the ip_queue kernel module loaded?" << endl;
      exit(1);
   }

   int status = ipq_set_mode(hIpq, IPQ_COPY_PACKET, sizeof(buf));
   if (status < 0)
   {
      *gLogStream << "ipq_set_mode failed; is the ip_queue kernel module loaded?" << endl;
      exit(1);
   }
   thread_response = pthread_create(&a_thread, NULL, talktoUI, NULL);
   if(thread_response != 0){
	perror("stats thread creation failed");
	exit(EXIT_FAILURE);
   }
   
   // Main IPQ loop, read and block packets
   while (!gStopBlockerThread)
   {
      ssize_t size = ipq_read(hIpq, buf, sizeof(buf), 1000000);
      if (size < 0)
      {
         *gLogStream << "ipq_read failed" << endl;
         exit(1);
      }
      else if (!size)
         continue;

      int type = ipq_message_type(buf);

      if (type == NLMSG_ERROR)
      {
         ipq_get_msgerr(buf);
         *gLogStream << "error: " << buf << endl;
         continue;
      }

      ipq_packet_msg_t* pkt = ipq_get_packet(buf);

      int verdict = NF_ACCEPT;

      if (pkt->data_len >= sizeof(struct iphdr))
      {
         struct iphdr* iphdr = (struct iphdr*)pkt->payload;
         Range find;

         // Check allowed ports
         if (iphdr->protocol == IPPROTO_TCP)
         {
            struct tcphdr* tcphdr = (struct tcphdr*)&pkt->payload[sizeof(struct iphdr)];
            if (gPrefs.allowedPorts[ntohs(tcphdr->source)] || gPrefs.allowedPorts[ntohs(tcphdr->dest)])
               goto ProcessPacket;
         }
         else if (iphdr->protocol == IPPROTO_UDP)
         {
            struct udphdr* udphdr = (struct udphdr*)&pkt->payload[sizeof(struct iphdr)];
            if (gPrefs.allowedPorts[ntohs(udphdr->source)] || gPrefs.allowedPorts[ntohs(udphdr->dest)])
               goto ProcessPacket;
         }

         // Check block list
         if (block_list->Find(&find, iphdr->saddr) || block_list->Find(&find, iphdr->daddr))
         {
            verdict = NF_DROP;
            unsigned char* s_ip = (unsigned char*)&iphdr->saddr;
            unsigned char* d_ip = (unsigned char*)&iphdr->daddr;
	    num_blocked++;
	    
            *gLogStream << "dropped src: "
               << (int)s_ip[0] << "." << (int)s_ip[1] << "." << (int)s_ip[2] << "." << (int)s_ip[3]
               << " dst: "
               << (int)d_ip[0] << "." << (int)d_ip[1] << "." << (int)d_ip[2] << "." << (int)d_ip[3]
               << " (" << find.rule << ")" << endl;

            (*block_list)++;
         }
	 else{
	    num_checked++;
	 }
      }

ProcessPacket:
      status = ipq_set_verdict(hIpq, pkt->packet_id, verdict, 0, NULL);
      if (status < 0)
      {
         *gLogStream << "ipq_set_verdict failed" << endl;
         exit(1);
      }
   }

   return 0;
}

} // namespace PG

#endif // !defined(__APPLE__) && !defined(__BSD__)

