//
//  peerguardnf (PeerGuard for Linux, BSD, and MacOS X systems)
//  Copyright (C) 2005 BitRot
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Id: HttpRequest.cpp,v 1.3 2005/03/27 01:26:18 nullcontext Exp $

#include "HttpServer.h"
#include "PeerGuard.h"
#include "sha1.h"
#include <sstream>

using namespace std;
using namespace __gnu_cxx;

namespace PG {

static const char VERSION[] = "0.4";
static hash_map<string,time_t,string_hash> gAuthorized;

bool UserIsAuthorized(const string& auth)
{
   hash_map<string,time_t,string_hash>::iterator itr = gAuthorized.find(auth);
   if (itr == gAuthorized.end())
      return false;

   if (time(0) > itr->second)
   {
      gAuthorized.erase(itr);
      return false;
   }

   return true;
}

void HandleHttpRequest(int sock, const string& req, BlockList* block_list)
{
   ostringstream out;
   CGIData data;

   GetURLData(&data, req);

   string auth = data["PGAUTH"];
   string name = data["UserName"];
   string pass = data["Password"];
   string cookie;

   if (!name.empty() && !pass.empty())
   {
      if (name == gPrefs.adminName && gPrefs.adminPassword == sha1(pass))
      {
         srand48(time(0));
         string val;
         for (int i = 0; i < 5; i++)
         {
            long r = mrand48();
            val.append((char*)&r, sizeof(r));
         }
         val.append(name);
         val.append(pass);
         auth = sha1(val);
         gAuthorized[auth] = time(0) + 900;
         cookie = "Set-Cookie: PGAUTH=" + auth + "; path=/;\r\n";
      }
   }

   out << "HTTP/1.1 200 OK\r\nServer: peerguardnf/" << VERSION << " (Unix)\r\n"
       << "X-Powered-By: You need to wind it\r\n"
       << "Cache-control: no-cache\r\n"  // FIXME: are these needed?
       << "Pragma: no-cache\r\n"         // FIXME: are these needed?
       << cookie
       << "Content-Type: text/html\r\n\r\n"
       << "<html><head></head><body>\r\n";

   if (!UserIsAuthorized(auth))
   {
      out << "<center>Please Login:<br><br>"
          << "<form method=\"get\">"
          << "<table border=\"0\">"
          << "<tr><td align=\"right\">Name:</td><td><input type=\"text\" name=\"UserName\"></td></tr>"
          << "<tr><td align=\"right\">Pass:</td><td><input type=\"text\" name=\"Password\"></td></tr>"
          << "<tr><td colspan=\"2\">&nbsp;</td></tr>"
          << "<tr><td colspan=\"2\" align=\"right\"><input type=\"submit\" value=\"Login\"></td></tr>"
          << "</table></form></center>";
   }
   else
   {
      out << "[<a href=\"/\">Main Menu</a>]<br><br>\r\n";

      string command = data["COMMAND"];
      
      if (command.empty())
      {
         // Default Menu
         out << "<a href=\"viewblocklist?\">View blocklist</a><br>";
         out << "<a href=\"viewstats?\">View stats</a><br>";
      }
      else if (command == "viewblocklist")
      {
         out << "This is the blocklist:<br>\r\n";
         WriteData(sock, out.str());
         out.str("");
         
         block_list->Lock();
         BlockTable::const_iterator itr = block_list->Table().begin();
         for (; itr != block_list->Table().end(); itr++)
         {
            if (WriteData(sock, itr->ToString() + "<br>\r\n"))
               break;
         }
            
         block_list->Unlock();
      }
      else if (command == "viewstats")
      {
         out << "Statistics:<br>\r\n"
             << block_list->RangeCount() << " ranges (" << block_list->IpCount() 
             << " IP addresses)<br>\r\n"
             << "Blocked packets: " << block_list->BlockCount() << "<br>\r\n";
      }
      else
      {
         // Invalid or bad request
         out.str("");
         out << "HTTP/1.1 400 Bad Request\r\nServer: peerguardnf/" << VERSION << " (Unix)\r\nConnection: close\r\nContent-Type: text/html\r\n\r\n";
         out << "<html><head></head><body>Invalid request";
      }
   }
   
   out << "</body></html>";
   
   WriteData(sock, out.str());
}

} // namespace PG

