//
//  peerguardnf (PeerGuard for Linux, BSD, and MacOS X systems)
//  Copyright (C) 2005 bitrot
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Id: Main.cpp,v 1.2 2005/03/27 01:28:26 nullcontext Exp $

#include "PeerGuard.h"
#include "HttpServer.h"
#include "sha1.h"
#include <unistd.h>
#include <signal.h>
#include <iostream>
#include <fstream>

using namespace PG;
using namespace std;

static bool gStopServer = false;

namespace PG {
int SafeExec(const char* cmd, bool output = true);
}

void usage(const char* prog)
{
   cerr << "usage: " << prog 
        << " [-d] [-f filename] [-p port,...] [-i address] [-l filename] [-c filename] [-h] [-s password]" << endl 
	<< "[-n blocklist.p2b] blocklist" 
        << endl << endl
        << "  -d : daemon mode" << endl
        << "  -f : set output filename (merge)" << endl
        << "  -p : list of ports to past through without blocking" << endl
        << "  -i : check IP address" << endl
        << "  -l : log filename (default stderr)" << endl
        << "  -c : config filename" << endl
        << "  -h : disable HTTP server" << endl
	<< "  -n : convert new blocklist format file to old blocklist format file" << endl
	<< "  -s : generate sha1 encrypted key for web server password" << endl
#if !defined(__APPLE__) && !defined(__BSD__)
        << "  -m : enable automatic kernel module loading" << endl
#endif
        << endl;
      
   exit(1);
}

int p2b_p2p(std::string filename) {
		char buf[7];
		unsigned char version;
		std::string outfile;
		std::ostringstream out;
		outfile=filename+".p2p";
		ifstream p2bfile(filename.c_str(), ios::binary);
		ofstream p2pfile(outfile.c_str(), ios::out);
		unsigned int start, end;
		
		string name;
		
		if(
			!p2bfile.read(buf, sizeof(buf)) ||
			memcmp(buf, "\xFF\xFF\xFF\xFFP2B", 7) ||
			!p2bfile.read((char*)&version, sizeof(version))
		) {
			cout << "wrong input blocklist format" << endl;
			p2pfile.close();
  			p2bfile.close();
			return 0;
		}

		if(version!=1 && version!=2)
			cout << "unknown p2b version";

		while(getline(p2bfile, name, '\0')) {
			if(
				!p2bfile.read((char*)&start, sizeof(start)) ||
				!p2bfile.read((char*)&end, sizeof(end))
			) 
			cout << "invalid p2b stream"; 
			uint32_t t_ip = ntohl(start);
      			char* addr = (char*)&t_ip;
			out.str("");
			out << name  << ":"
      			    << (unsigned int)(addr[3] & 0xFF) << "." 
          		    << (unsigned int)(addr[2] & 0xFF) << "." 
           		    << (unsigned int)(addr[1] & 0xFF) << "." 
          		    << (unsigned int)(addr[0] & 0xFF) << "-";
			t_ip = ntohl(end);
      			addr = (char*)&t_ip;
			out << (unsigned int)(addr[3] & 0xFF) << "." 
          		    << (unsigned int)(addr[2] & 0xFF) << "." 
           		    << (unsigned int)(addr[1] & 0xFF) << "." 
          		    << (unsigned int)(addr[0] & 0xFF) << endl;
			p2pfile << out.str();
		}
		cout << "file saved as " << filename << ".p2p" << endl;
		p2pfile.close();
		p2bfile.close();
		
		return 0;
}
void SignalHandler(int sig)
{
   switch (sig)
   {
   case SIGINT:
   case SIGTERM:
      if (!gStopServer)
      {
         *gLogStream << "caught signal, cleaning up and terminating" << endl;
         gStopServer = true;
      }
      break;
   default:
      break;
   }
}

void ExitHandler(void)
{
   // Very important to remove the rules at exit, otherwise the machine may
   // not be able to access the network.
   DeleteFirewallRules();
}

extern "C"
int main(int argc, char* argv[])
{
   gLogStream = &cerr;
   ofstream log_file;
   string output_file;
   string config_file;
   uint32_t check_ip = 0;
   bool daemon = false;
   bool disable_http = false;
   bool p2b = false;
   
   if (argc < 2)
      usage(argv[0]);
   else if (getuid())
   {
      cerr << "This program must be run as root" << endl;
      exit(1);
   }

   int arg_char;
   while((arg_char = getopt(argc, argv, "mhdf:p:i:l:s:c:n:")) != -1)
   {
      switch (arg_char)
      {
      case 'd':
         daemon = true;
         break;
      case 'f':
	 output_file = optarg;
         break;
      case 'p':
         {
            char* temp;
            char* buf = new char[strlen(optarg) + 1];
            strcpy(buf, optarg);
            char* pos = strtok_r(buf, ", ", &temp);
            while(pos)
            {
               gPrefs.AllowPort(atoi(pos));
               pos = strtok_r(0, ", ", &temp);
            }

            delete [] buf;
         }
         break;
      case 'i':
         {
            int ip1_1, ip1_2, ip1_3, ip1_4;
            int count = sscanf(optarg, "%d.%d.%d.%d", &ip1_1, &ip1_2, &ip1_3, &ip1_4);
            if (count == 4)
            {
               char* pos = (char*)&check_ip;
               pos[0] = (ip1_1 & 0xFF);
               pos[1] = (ip1_2 & 0xFF);
               pos[2] = (ip1_3 & 0xFF);
               pos[3] = (ip1_4 & 0xFF);
            }
            else
            {
               cerr << "Invalid IP address specified" << endl;
               exit(1);
            }
         }
         break;
      case 'l':
         log_file.open(optarg, ofstream::out | ofstream::app);
         if (!log_file.is_open())
            cerr << "Failed to open log file \"" << optarg << "\", using STDERR" << endl;
         else
            gLogStream = &log_file;
         break;
      case 'c':
         config_file = optarg;
         break;
      case 'h':
         disable_http = true;
         break;
      case 's':
	 cout << sha1(optarg) << endl;
	 exit(0);
	 break;
      case 'n':
         p2b = true;
	 gPrefs.blocklistFilename = optarg;
         break;
#if !defined(__APPLE__) && !defined(__BSD__)
      case 'm':
         PG::SafeExec("modprobe ip_tables", false);
         usleep(2000);
         PG::SafeExec("modprobe iptable_filter", false);
         usleep(2000);
         PG::SafeExec("modprobe ip_queue", false);
         usleep(2000);
         break;
#endif
      default:
         usage(argv[0]);
         break;
      }
   }

   if (!config_file.empty())
      gPrefs.Read(config_file);

   // Override the blocklist filename if specified on command line
   if (optind < argc)
      gPrefs.blocklistFilename = argv[optind];

   BlockList list;

   if (!gPrefs.blocklistFilename.empty())
      *gLogStream << "Reading blocklist" << endl;
   else
   {
      gPrefs.blocklistFilename = "-";
      *gLogStream << "Reading blocklist from STDIN" << endl;
   }
   
   if(p2b == true){
   	p2b_p2p(gPrefs.blocklistFilename);
  	exit(0);
   }
   

   if (list.ReadFile(gPrefs.blocklistFilename))
   {
      *gLogStream << "Failed to read blocklist" << endl;
      return 1;
   }

   if (check_ip)
   {
      if (list.Find(0, check_ip))
         *gLogStream << "IP address is in blocklist; Address will be blocked" << endl;
      else
         *gLogStream << "IP address will not be blocked" << endl;

      return 0;
   }

   if (!output_file.empty())
   {
      if (list.SaveFile(output_file))
      {
         *gLogStream << "Failed to write merged blocklist" << endl;
         return 0;
      }

      *gLogStream << "Successfully wrote merged list to \"" << output_file << "\"" << endl;

      return 0;
   }

   signal(SIGINT, SignalHandler);
   signal(SIGTERM, SignalHandler);
   signal(SIGPIPE, SignalHandler);

   if (daemon)
   {
      pid_t pid = fork();
      if (pid == -1)
         return -1;
      else if (pid)
         return 0;
      else
      {
         *gLogStream << "Entering daemon mode" << endl;
         close(0);
         close(1);
         close(2);
      }
   }

   // Start blocking packets
   if (!StartBlockerThread(&list))
   {
      CreateFirewallRules();
      atexit(ExitHandler);

      if (!disable_http)
         StartHttpServer(&list);

      // Main loop
      while (!gStopServer)
         usleep(1000);
         
      if (!disable_http)
         StopHttpServer();

      StopBlockerThread();
   }

   return 0;
}

